use color_eyre::Result;
use ratatui::{
    layout::{Alignment, Rect},
    text::Line,
    widgets::{Block, Paragraph, Wrap},
    Frame,
};

use super::Component;
use crate::{
    action::Action, palette::StylePalette, widgets::OghamCenter
};

#[derive(Debug, Clone, Default)]
pub struct Version {
    genesis_hash: Option<String>,
    node_version: Option<String>,
    chain_name: Option<String>,
    palette: StylePalette,
}

impl Version {
    fn set_chain_name(&mut self, chain_name: Option<String>) -> Result<()> {
        self.chain_name = chain_name;
        Ok(())
    }

    fn set_node_version(&mut self, node_version: Option<String>) -> Result<()> {
        self.node_version = node_version;
        Ok(())
    }

    fn set_genesis_hash(&mut self, genesis_hash: Option<String>) -> Result<()> {
        self.genesis_hash = genesis_hash;
        Ok(())
    }

    fn prepared_genesis_hash(&self) -> String {
        if self.genesis_hash.is_some() {
            let genesis_hash = self.genesis_hash.clone().unwrap();
            let len = genesis_hash.len();
            format!("Genesis: {}...{}", &genesis_hash[0..4], &genesis_hash[len-6..])
        } else {
            OghamCenter::default().to_string()
        }
    }
}

impl Component for Version {
    fn update(&mut self, action: Action) -> Result<Option<Action>> {
        match action {
            Action::SetChainName(name) => self.set_chain_name(name)?,
            Action::SetNodeVersion(version) => self.set_node_version(version)?,
            Action::SetGenesisHash(genesis) => self.set_genesis_hash(genesis)?,
            _ => {}
        };
        Ok(None)
    }

    fn draw(&mut self, frame: &mut Frame, area: Rect) -> Result<()> {
        let [_, version] = super::menu_layout(area);

        let text_style = self.palette.create_text_style(false);
        let (border_style, border_type) = self.palette.create_border_style(false);
        let text = vec![
            Line::styled(self.chain_name.clone().unwrap_or(OghamCenter::default().to_string()), text_style),
            Line::styled(self.node_version.clone().unwrap_or(OghamCenter::default().to_string()), text_style),
            Line::styled(self.prepared_genesis_hash(), text_style),
        ];
        let paragraph = Paragraph::new(text)
            .block(Block::bordered()
                .border_style(border_style)
                .border_type(border_type)
            )
            .alignment(Alignment::Center)
            .wrap(Wrap { trim: true });

        frame.render_widget(paragraph, version);
        Ok(())
    }
}
