use std::sync::Arc;
use tokio::sync::mpsc::UnboundedSender;
use once_cell::sync::Lazy;
use reqwest::Client;
use color_eyre::Result;
use rand::RngCore;
use serde::{Deserialize, de::DeserializeOwned};

use crate::{action::Action, types::params::RpcParams};

mod active_era;
mod health_check;
mod node_name;
mod genesis_hash;
mod chain_name;
mod node_version;
mod latest_block;
mod finalized_block;
mod current_epoch;
mod validators;
mod tx_pool;

pub use active_era::ActiveEraRequest;
pub use health_check::HealthCheckRequest;
pub use node_name::NodeNameRequest;
pub use genesis_hash::GenesisHashRequest;
pub use chain_name::ChainNameRequest;
pub use node_version::NodeVersionRequest;
pub use latest_block::LatestBlockRequest;
pub use finalized_block::FinalizedBlockRequest;
pub use current_epoch::CurrentEpochRequest;
pub use validators::ValidatorsRequest;
pub use tx_pool::TxPoolRequest;

static CLIENT: Lazy<Arc<Client>> = Lazy::new(|| Arc::new(Client::new()));
const DEFAULT_URL: &str = "http://localhost:9945";

pub type AppActionSender = UnboundedSender<Action>;

#[derive(Debug, Deserialize)]
pub struct GhostResponse<ResponseType> {
    result: ResponseType,
}

#[derive(Default)]
struct GhostRequestBuilder<'a> {
    action_tx: Option<AppActionSender>,
    id: u32,
    url: &'a str,
    timeout: std::time::Duration,
}

impl<'a> GhostRequestBuilder<'a> {
    pub fn with_id(mut self, id: u32) -> Self {
        self.id = id;
        self
    }

    pub fn with_url(mut self, url: &'a str) -> Self {
        self.url = url;
        self
    }

    pub fn with_timeout(mut self, timeout: std::time::Duration) -> Self {
        self.timeout = timeout;
        self
    }

    pub fn with_action_tx(mut self, action_tx: AppActionSender) -> Self {
        self.action_tx = Some(action_tx);
        self
    }

    pub fn build(self) -> GhostRequest<'a> {
        GhostRequest {
            action_tx: self.action_tx.expect("channel sender should exist"),
            id: self.id,
            url: self.url,
            timeout: self.timeout,
        }
    }
}

#[derive(Debug)]
pub struct GhostRequest<'a> {
    action_tx: AppActionSender,
    id: u32,
    url: &'a str,
    timeout: std::time::Duration,
}

impl<'a> GhostRequest<'a> {
    pub async fn send<ResponseType>(
        &self,
        method: &str,
        params: RpcParams,
    ) -> Result<GhostResponse<ResponseType>>
    where
        ResponseType: DeserializeOwned,
    {
        Ok(CLIENT
            .post(self.url)
            .header(reqwest::header::CONTENT_TYPE, "application/json")
            .body(format!("{{\"id\":{},\"jsonrpc\":\"2.0\",\"method\":{:?},\"params\":{}}}",
                    self.id, method, params.build()))
            .timeout(self.timeout)
            .send()
            .await?
            .json::<GhostResponse<ResponseType>>()
            .await?
        )
    }
}

pub struct Network<'a> {
    action_tx: AppActionSender,
    timeout: std::time::Duration,
    internal_randomness: rand::rngs::ThreadRng,
    url: &'a str,
}

impl<'a> Network<'a> {
    pub fn new(action_tx: AppActionSender) -> Self {
        Self { 
            action_tx,
            timeout: Default::default(),
            internal_randomness: rand::thread_rng(),
            url: DEFAULT_URL,
        }
    }

    pub fn with_url(mut self, url: &'a str) -> Self {
        self.url = url;
        self
    }

    pub fn with_timeout(mut self, timeout: u64) -> Self {
        self.timeout = std::time::Duration::from_secs(timeout);
        self
    }

    pub async fn handle_network_event(&mut self, io_event: Action) -> Result<()> {
        let request = GhostRequestBuilder::default()
            .with_action_tx(self.action_tx.clone())
            .with_id(self.internal_randomness.next_u32())
            .with_url(self.url)
            .with_timeout(self.timeout)
            .build();

        match io_event {
            Action::GetSyncState => HealthCheckRequest(request).send().await,
            Action::GetNodeName => NodeNameRequest(request).send().await,
            Action::GetGenesisHash => GenesisHashRequest(request).send().await,
            Action::GetChainName => ChainNameRequest(request).send().await,
            Action::GetNodeVersion => NodeVersionRequest(request).send().await,
            Action::GetLatestBlock => LatestBlockRequest(request).send().await,
            Action::GetFinalizedBlock => FinalizedBlockRequest(request).send().await,
            Action::GetActiveEra => ActiveEraRequest(request).send().await,
            Action::GetEpoch => CurrentEpochRequest(request).send().await,
            Action::GetValidators => ValidatorsRequest(request).send().await,
            Action::GetPendingExtrinsics => TxPoolRequest(request).send().await,
            _ => Ok(())
        }
    }
}
