use color_eyre::Result;
use serde::Serialize;

#[derive(Debug)]
pub struct RpcParams(ParamsBuilder);

impl RpcParams {
    pub fn new() -> Self {
        Self::default()
    }

    pub fn insert<P: Serialize>(&mut self, value: P) -> Result<()> {
        self.0.insert(value)
    }

    pub fn build(self) -> String {
        self.0.build()
    }
}

impl Default for RpcParams {
    fn default() -> Self {
        Self(ParamsBuilder::positional())
    }
}

const PARAM_BYTES_CAPACITY: usize = 128;

#[derive(Debug)]
pub struct ParamsBuilder {
    bytes: Vec<u8>,
    start: char,
    end: char,
}

impl ParamsBuilder {
    fn new(start: char, end: char) -> Self {
        Self { bytes: Vec::new(), start, end }
    }

    fn positional() -> Self {
        Self::new('[', ']')
    }

    fn maybe_initialize(&mut self) {
        if self.bytes.is_empty() {
            self.bytes.reserve(PARAM_BYTES_CAPACITY);
            self.bytes.push(self.start as u8);
        }
    }

    pub fn build(mut self) -> String {
        if self.bytes.is_empty() {
            return format!("{}{}", self.start, self.end);
        }

        let index = self.bytes.len() - 1;
        if self.bytes[index] == b',' {
            self.bytes[index] = self.end as u8;
        } else {
            self.bytes.push(self.end as u8);
        }

        unsafe { String::from_utf8_unchecked(self.bytes) }
    }

    pub fn insert<P: Serialize>(&mut self, value: P) -> Result<()> {
        self.maybe_initialize();

        serde_json::to_writer(&mut self.bytes, &value)?;
        self.bytes.push(b',');

        Ok(())
    }
}
